@@ -2,7 +2,7 @@
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TypeFamilies       #-}
 
-module Main (main) where
+module KeyValueNoTH (main) where
 
 import           Data.Acid
 import           Data.Acid.Advanced
@@ -93,6 +93,6 @@ instance Method LookupKey where
 instance QueryEvent LookupKey
 
 instance IsAcidic KeyValue where
-    acidEvents = [ UpdateEvent (\(InsertKey key value) -> insertKey key value)
-                 , QueryEvent (\(LookupKey key) -> lookupKey key)
+    acidEvents = [ UpdateEvent (\(InsertKey key value) -> insertKey key value) safeCopyMethodSerialiser
+                 , QueryEvent (\(LookupKey key) -> lookupKey key)              safeCopyMethodSerialiser
                  ]
