@@ -2,14 +2,19 @@
 {-# LANGUAGE TemplateHaskell    #-}
 {-# LANGUAGE TypeFamilies       #-}
 
-module Main (main) where
+-- With optimizations enabled, serializing the checkpoint can happen too quickly
+{-# OPTIONS_GHC -O0 #-}
+
+module SlowCheckpoint (main) where
 
 import           Data.Acid
 
 import           Control.Concurrent
+import           Control.Monad.Reader
 import           Control.Monad.State
 import           Data.SafeCopy
 import           Data.Time
+import           System.Directory
 import           System.IO
 
 ------------------------------------------------------
@@ -36,13 +41,20 @@ tick = do SlowCheckpoint slow tick <- get
           put $ SlowCheckpoint slow (tick+1)
           return tick
 
-$(makeAcidic ''SlowCheckpoint ['setComputationallyHeavyData, 'tick])
+askTick :: Query SlowCheckpoint Int
+askTick = do SlowCheckpoint _ tick <- ask
+             return tick
+
+$(makeAcidic ''SlowCheckpoint ['setComputationallyHeavyData, 'tick, 'askTick])
 
 ------------------------------------------------------
 -- This is how AcidState is used:
 
 main :: IO ()
-main = do acid <- openLocalStateFrom "state/SlowCheckpoint" (SlowCheckpoint 0 0)
+main = do putStrLn "SlowCheckpoint test"
+          exists <- doesDirectoryExist fp
+          when exists $ removeDirectoryRecursive fp
+          acid <- openLocalStateFrom fp (SlowCheckpoint 0 0)
           putStrLn "This example illustrates that the state is still accessible while"
           putStrLn "a checkpoint is being serialized. This is an important property when"
           putStrLn "the size of a checkpoint reaches several hundred megabytes."
@@ -51,12 +63,20 @@ main = do acid <- openLocalStateFrom "state/SlowCheckpoint" (SlowCheckpoint 0 0)
           putStrLn ""
           doTick acid
           update acid SetComputationallyHeavyData
-          forkIO $ do putStrLn "Seriazing checkpoint..."
+          forkIO $ do putStrLn "Serializing checkpoint..."
                       t <- timeIt $ createCheckpoint acid
-                      putStrLn $ "Checkpoint created in: " ++ show t
+                      n <- query acid AskTick
+                      putStrLn $ "Checkpoint created in: " ++ show t ++ " (saw " ++ show n ++ " ticks)"
+                      when (n < threshold) $ error $ "Not enough ticks! Expected at least " ++ show threshold
           replicateM_ 20 $
             do doTick acid
                threadDelay (10^5)
+          putStrLn "SlowCheckpoint done"
+  where
+    fp = "state/SlowCheckpoint"
+
+    -- We must see at least this many ticks for the test to be considered a success
+    threshold = 5
 
 doTick acid
     = do tick <- update acid Tick
