@@ -2,7 +2,7 @@
 {-# LANGUAGE StandaloneDeriving #-}
 {-# LANGUAGE TypeFamilies       #-}
 
-module Main (main) where
+module StressTestNoTH (main) where
 
 import           Data.Acid
 import           Data.Acid.Advanced
@@ -84,6 +84,6 @@ instance Method QueryState where
 instance QueryEvent QueryState
 
 instance IsAcidic StressState where
-    acidEvents = [ UpdateEvent (\PokeState -> pokeState)
-                 , QueryEvent (\QueryState -> queryState)
+    acidEvents = [ UpdateEvent (\PokeState -> pokeState)  safeCopyMethodSerialiser
+                 , QueryEvent (\QueryState -> queryState) safeCopyMethodSerialiser
                  ]
