@@ -1,19 +1,16 @@
-{-# LANGUAGE DeriveDataTypeable #-}
 {-# LANGUAGE TemplateHaskell    #-}
 {-# LANGUAGE TypeFamilies       #-}
 
-module Main (main) where
+module ChangeState (main, test) where
 
 import           Data.Acid
 
+import           Control.Exception
 import           Control.Monad.State
 import           Data.SafeCopy
+import           System.Directory
 import           System.Environment
-
-import           Data.Typeable
-
-import           Control.Exception
-import           Prelude             hiding (catch)
+import           Data.List (isSuffixOf)
 
 import qualified Data.Text           as Text
 
@@ -43,9 +40,24 @@ main = do putStrLn "This example simulates what happens when you modify your sta
           putStrLn "without telling AcidState how to migrate from the old version to the new."
           putStrLn "Hopefully this program will fail with a readable error message."
           putStrLn ""
-          firstAcid <- openLocalStateFrom "state/ChangeState" (FirstState "first state")
+          firstAcid <- openLocalStateFrom fp (FirstState "first state")
           createCheckpoint firstAcid
           closeAcidState firstAcid
-          secondAcid <- openLocalStateFrom "state/ChangeState" (SecondState (Text.pack "This initial value shouldn't be used"))
+          secondAcid <- openLocalStateFrom fp (SecondState (Text.pack "This initial value shouldn't be used"))
           closeAcidState secondAcid
-          putStrLn "If you see this message then something has gone wrong!"
+          error "If you see this message then something has gone wrong!"
+
+test :: IO ()
+test = do
+    putStrLn "ChangeState test"
+    exists <- doesDirectoryExist fp
+    when exists $ removeDirectoryRecursive fp
+    handle hdl main
+    putStrLn "ChangeState done"
+  where
+    hdl (ErrorCall msg)
+      | "Could not parse saved checkpoint due to the following error: too few bytes\nFrom:\tChangeState.SecondState:\n\tdemandInput\n\n" `isSuffixOf` msg 
+      = putStrLn $ "Caught error: " ++ msg
+    hdl e = throwIO e
+
+fp = "state/ChangeState"
